//fileio.h
//Utilities for reading and writing to files in the game's subdirectories


//fileio works by creating a table of all files within the applications directory
//at startup. this way all files we need can be addressed simply using reference numbers.
//sometimes also refered to as "ID numbers".

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "fileio.h"
#include "platform.h"
#include "gamemem.h"
#include "error.h"
#include "parser.h"

int gFileTableSize;				//size of the file reference table
int gFileTableExtendedSize;
tFileTableEntry *gFileTable;	//the file reference table

void FileInitFileTable(tFileTableEntry *fileTable,int maxSize,int *fileTableSize,int reInit);

//Initialize file reference table
void FileInitIO()
{
	gFileTableExtendedSize=0;
	gFileTableSize=0;
	gFileTable=(tFileTableEntry*)calloc(kMaxFiles,sizeof(tFileTableEntry));
	FileInitFileTable(gFileTable,kMaxFiles,&gFileTableSize,false);
}

void FileRescanDirectory()
{
	int totalSize=gFileTableSize+gFileTableExtendedSize;
	FileInitFileTable(gFileTable,kMaxFiles,&totalSize,true);
	gFileTableExtendedSize=totalSize-gFileTableSize;
}

void FileReleaseData(tFileRef reference)
{
	if(gFileTable[reference].loaded)
	{
		MemoryFreeBlock(gFileTable[reference].data);
		gFileTable[reference].loaded=false;
	}
}

//Compare two File Table entries for sorting and searching
int FileTableCompare(const void *a,const void *b)
{
	return _stricmp(((tFileTableEntry*)a)->name,((tFileTableEntry*)b)->name);
}

int gFileErrorReporting=true;

//Search file reference table for a file matching a name and return its reference number
tFileRef FileGetReference(char *name)
{
	if(*name=='\0')
		return -1;
	//the key to search for	
	tFileTableEntry key;
#ifndef __TARGET_TEXTURECOMPRESSOR
	snprintf(key.name,32,"%s.txr",name);
	
	//search for an entry matching our search key
	tFileTableEntry *result=(tFileTableEntry*)bsearch(&key,gFileTable,gFileTableSize,sizeof(tFileTableEntry),&FileTableCompare);
	
	//return results
	if(result)
		return result-gFileTable;
	else
	{
#endif
		strcpy(key.name,name);
		tFileTableEntry *result=(tFileTableEntry*)bsearch(&key,gFileTable,gFileTableSize,sizeof(tFileTableEntry),&FileTableCompare);
		if(result)
			return result-gFileTable;
		else
		{
			for(int i=gFileTableSize;i<gFileTableSize+gFileTableExtendedSize;i++)
				if(_stricmp(gFileTable[i].name,name)==0)	
					return i; 	
			if(gFileErrorReporting)
			{
				//or give an error message
				PrintConsoleString("File not found!: %s\n",name);
			}
			return -1;
		}
#ifndef __TARGET_TEXTURECOMPRESSOR
	}
#endif
}

tFileRef FileGetIndexedReference(tFileRef base,int i)
{
	if(i>0)
	{
		char nameCopy[kMaxFileNameLength];
		strcpy(nameCopy,FileGetName(base));
		char *found=strchr(nameCopy,'.');

		char indexedName[kMaxFileNameLength];
		if(found)
		{
			*found='\0';
			sprintf(indexedName,"%s#%x.%s",nameCopy,i,found+1);
		}
		else
			sprintf(indexedName,"%s#%x",nameCopy,i);
		
		int seek=base-1;
		while(_stricmp(gFileTable[seek].name,indexedName)>0&&seek>0)
			seek--;
		
		if(_stricmp(gFileTable[seek].name,indexedName))
			return base;
		else
			return seek;
	}
	else
		return base;
}

tFileRef FileGetBaseReference(tFileRef ref)
{
	char *found=strchr(FileGetName(ref),'.');
	if(!found)
		return ref;
	
	char *cross=found;
	while(*cross!='#'&&cross>FileGetName(ref))
		cross--;
	if(*cross!='#')
		return ref;

	int l=cross-FileGetName(ref);
	char nameCopy[kMaxFileNameLength];
	memcpy(nameCopy,FileGetName(ref),l);
	strcpy(nameCopy+l,found);
	tFileRef base=FileGetReference(nameCopy);
	return base!=kFileErr?base:ref;	
}

//platform specific, can be found in macfileio.cpp
int FileLoadData(tFileRef fileRef);
void FileStoreData(tFileRef fileRef);

//return the extension of a file given it's reference number
char *FileGetExtension(tFileRef reference)
{
	char *found=NULL;
	char *dot=gFileTable[reference].name;
	do{
		dot=strchr(dot,'.');
		if(dot){
			found=dot;
			dot++;
		}
	}while(dot);
	if(found)
		return found;
	else 
		return NULL;//'\0';
}

//return the name of a file given it's reference number
char *FileGetName(tFileRef reference)
{
	return gFileTable[reference].name;
}

//return the size of a file given it's reference number
int FileGetSize(tFileRef reference)
{
	if(!gFileTable[reference].loaded)
		gFileTable[reference].loaded=FileLoadData(reference);
	return gFileTable[reference].size;
}

//return a pointer to the file's data (and load the file if necessary)
void *FileGetDataPtr(tFileRef reference)
{
	if(!gFileTable[reference].loaded)
		gFileTable[reference].loaded=FileLoadData(reference);
	return gFileTable[reference].loaded?gFileTable[reference].data:(void*)0;
}

unsigned int FileGetChecksum(tFileRef reference)
{
	return MemoryChecksum(FileGetDataPtr(reference),FileGetSize(reference));
}

//Changes a file's contents to data, and writes the file to disk.
void FileSetData(tFileRef reference,void *data)
{
	if(gFileTable[reference].loaded&&data!=gFileTable[reference].data)
		MemoryFreeBlock(gFileTable[reference].data);
	gFileTable[reference].data=data;
	gFileTable[reference].loaded=true;
	gFileTable[reference].size=MemoryBlockSize(data);
	FileStoreData(reference);
}

//parses a file's data using a file format parser 
//and return a pointer to parsed data structure
void* FileGetParsedDataPtr(tFileRef reference, int parserType, int dataSize)
{
	if(!gFileTable[reference].parsed)
	{
		gFileTable[reference].parsedData=MemoryAllocateZeroedBlock(dataSize);
		ParseFile(reference,gFileTable[reference].parsedData,parserType);
		gFileTable[reference].parsed=true;
	}
	return gFileTable[reference].parsedData;		
}